#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <GL/glew.h>
#include <GL/glu.h>

#include <SDL.h>
#include <SDL_opengl.h>
#include <SDL_image.h>

#include "mathext.h"
#include "shaders.h"
#include "render_text.h"


#define SCREEN_WIDTH        640
#define SCREEN_HEIGHT       480

#define POSITION            0
#define NORMAL              1

#define FRAME_LAYERS        16


SDL_Window* gWindow = NULL;
SDL_GLContext gContext;
SDL_Surface *image2d;
unsigned int last_time;

#define MOTION_BLUR_SHADING             0
#define POSTPROCESSING_RECT             1
#define PROGRAM_COUNT                   2

GLuint shader_program[PROGRAM_COUNT];
GLuint a_position_loc;
GLuint a_texture_coords_loc;
GLuint u_mvpMatrix_loc;

#define POSITION                        0
#define TEXTURE_COORD                   1
#define TEX_COORD                       2
#define POSITION_RECT                   2
#define TEXCOORD_RECT                   3
#define VERTEX_BUFFER_COUNT             4

#define PLANE                           0
#define RECTANGLE                       1
#define VERTEX_ARRAY_COUNT              2

GLuint vertex_array[VERTEX_ARRAY_COUNT];
GLuint vertex_buffer[VERTEX_BUFFER_COUNT];
GLuint vertex_indices_buffer;
GLuint texture;

float projection_matrix[16];
float modelview_matrix[FRAME_LAYERS][16];
float mvp_matrix[16];
float aux_matrix1[16];
float aux_matrix2[16];

GLfloat left = -2.0f;
GLfloat right = 2.0f;
GLfloat bottom = -2.0f;
GLfloat top = 2.0f;
GLfloat near = 3.0f;
GLfloat far = 7.0f;

GLfloat translate_x = 0.0f;
GLfloat translate_y = 0.0f;
GLfloat translate_z = 0.0f;


GLfloat fighter_plane[4*3] = {
    1.0f, 1.0f, 1.0f,
    -1.0f, 1.0f, 1.0f,
    -1.0f, -1.0f, 1.0f,
    1.0f, -1.0f, 1.0f
};
GLfloat tex_coord_fighter[4*2] = {
    1.0f, 0.0f,
    0.0f, 0.0f,
    0.0f, 1.0f,
    1.0f, 1.0f
};
GLuint indices[12*3] = {
    0, 1, 3,
    1, 2, 3,
};
GLfloat position_rect[4*2] = {
    -1.0f, -1.0f,
    1.0f, -1.0f,
    -1.0f, 1.0f,
    1.0f, 1.0f
};
GLfloat tex_coord_rect[4*2] = {
    0.0f, 0.0f,
    1.0f, 0.0f,
    0.0f, 1.0f,
    1.0f, 1.0f
};

GLuint frameBuffer;
#define FRAME_WIDTH             4*256
#define FRAME_HEIGHT            4*256
#define COLOR0                  0
#define DEPTH                   1
#define RENDER_TEXTURE_COUNT    2
GLuint renderTexture[RENDER_TEXTURE_COUNT];


int init();
int initGL();
int check_gl_error();
void cleanup();
int set_shaders();
int init_buffers();
void update();
void render();

char *text_file_read(char *fname);




int init() {
    // Initialize SDL
    if (SDL_Init(SDL_INIT_VIDEO) < 0) {
        printf("Błąd podczas inicjalizacji silnika SDL2: %s\n", SDL_GetError());
        return 0;
    }

    SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 1);
    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
    SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);

    gWindow = SDL_CreateWindow("Rozmycie ruchu (Motion blur)", SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED,
                                SCREEN_WIDTH, SCREEN_HEIGHT, SDL_WINDOW_OPENGL | SDL_WINDOW_SHOWN);
    if (gWindow == NULL) {
        printf("Błąd podczas tworzenia okna: %s\n", SDL_GetError());
        return 0;
    }

    gContext = SDL_GL_CreateContext(gWindow);
    if (gContext == NULL) {
        printf("Błąd podczas tworzenia kontekstu OpenGL: %s\n", SDL_GetError());
        return 0;
    }

    if (SDL_GL_SetSwapInterval(1) < 0) {
        printf( "Błąd podczas ustawiania VSync: %s\n", SDL_GetError());
    }

    int imgFlags = IMG_INIT_PNG | IMG_INIT_JPG;
    if (!(IMG_Init(imgFlags) & imgFlags)) {
        printf("Błąd podczas inicjalizacji biblioteki SDL_image: %s\n", IMG_GetError());
        return 0;
    }

    if (!initGL()) {
        printf("Błąd podczas inicjalizacji OpenGL!\n");
        return 0;
    }

    if (!init_draw_text()) {
        printf("Error in function init_draw_text\n");
        return 0;
    }

    last_time = SDL_GetTicks();
    return 1;
}

int initGL() {
    GLenum error = GL_NO_ERROR;
    GLfloat aspect = 1.0f;
    int i;

    glewExperimental = GL_TRUE;
    error = glewInit();
    while (check_gl_error() == GL_INVALID_ENUM) {};
    if (error != GLEW_OK) {
        printf("glewInit failed: %s\n", glewGetErrorString(error));
        return 0;
    }

    glClearColor(1.0f, 1.0f, 1.0f, 1.0f);


    aspect = (float)SCREEN_WIDTH / (float)SCREEN_HEIGHT;
    get_frustum_matrix(projection_matrix, left*aspect, right*aspect, bottom, top, near, far);

    glGenTextures(1, &texture);
    glBindTexture(GL_TEXTURE_2D, texture);
    image2d = IMG_Load("fighter.png");
    if (image2d == NULL) {
        printf("Błąd podczas ładowania pliku tekstury fighter.png: %s\n", SDL_GetError());
        return 0;
    }
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, image2d->w, image2d->h, 0, GL_RGBA, GL_UNSIGNED_BYTE, image2d->pixels);
    SDL_FreeSurface(image2d);
    image2d = NULL;
    glGenerateMipmap(GL_TEXTURE_2D);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glBindTexture(GL_TEXTURE_2D, 0);

    // utworzenie obiektów tekstur - na dane bufora koloru i bufora głębokości
    glGenTextures(RENDER_TEXTURE_COUNT, renderTexture);

    glBindTexture(GL_TEXTURE_2D_ARRAY, renderTexture[COLOR0]);
    glTexImage3D(GL_TEXTURE_2D_ARRAY, 0, GL_RGBA, FRAME_WIDTH, FRAME_HEIGHT, FRAME_LAYERS, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
    glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glBindTexture(GL_TEXTURE_2D_ARRAY, renderTexture[DEPTH]);
    glTexImage3D(GL_TEXTURE_2D_ARRAY, 0, GL_DEPTH_COMPONENT, FRAME_WIDTH, FRAME_HEIGHT, FRAME_LAYERS, 0, GL_DEPTH_COMPONENT, GL_FLOAT, NULL);
    glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_MIN_FILTER, GL_LINEAR);


    // utworzenie obiektu bufora ramki i dołączenie obiektów tekstury
    glGenFramebuffers(1, &frameBuffer);
    glBindFramebuffer(GL_DRAW_FRAMEBUFFER, frameBuffer);
    glFramebufferTexture(GL_DRAW_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, renderTexture[COLOR0], 0);
    glFramebufferTexture(GL_DRAW_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, renderTexture[DEPTH], 0);

    // sprawdzenie poprawności i kompletności obiektu bufora ramki
    error = glCheckFramebufferStatus(GL_DRAW_FRAMEBUFFER);
    if (error != GL_FRAMEBUFFER_COMPLETE ) {
        printf("Niepoprawny obiekt bufora ramki\n");
        exit(0);
    }

    // obszar renderingu - cała powierzchnia bufora ramki
    glViewportIndexedf(1, 0.0, 0.0, (GLfloat)FRAME_WIDTH, (GLfloat)FRAME_HEIGHT);



    if (!set_shaders()) {
        cleanup();
        return 0;
    }
    if (!init_buffers()) {
        cleanup();
        return 0;
    }


    // załadowanie początkowych ustawień tablicy z macierzami modelu-widoku
    get_identity_matrix(modelview_matrix[0]);
    get_translation_matrix4x4(aux_matrix1, 0.0f, 0.0f, -(near+far)/2.0f);
    mult_matrices(aux_matrix2, modelview_matrix[0], aux_matrix1, 4);
    // przesunięcie obiektu
    get_translation_matrix4x4(aux_matrix1, translate_x, translate_y, 0.0f);
    mult_matrices(modelview_matrix[0], aux_matrix2, aux_matrix1, 4);
    memcpy(modelview_matrix[0], aux_matrix2, sizeof(float) * 16);
    for (int layer = 1; layer < FRAME_LAYERS; layer++) {
        memcpy(modelview_matrix[layer], modelview_matrix[0], sizeof(float) * 16);
    }

    glEnable(GL_DEPTH_TEST);

    // obszar renderingu - całe okno
    glViewportIndexedf(0, 0.0, 0.0, (GLfloat)SCREEN_WIDTH, (GLfloat)SCREEN_HEIGHT);

    return 1;
}


int set_shaders() {
    shader_program[MOTION_BLUR_SHADING] = glCreateProgram();
    glAttachShader(shader_program[MOTION_BLUR_SHADING], load_shader(GL_VERTEX_SHADER, "motion_blur_vs.glsl"));
    glAttachShader(shader_program[MOTION_BLUR_SHADING], load_shader(GL_GEOMETRY_SHADER, "motion_blur_gs.glsl"));
    glAttachShader(shader_program[MOTION_BLUR_SHADING], load_shader(GL_FRAGMENT_SHADER, "motion_blur_fs.glsl"));
    link_and_validate_program(shader_program[MOTION_BLUR_SHADING]);

    shader_program[POSTPROCESSING_RECT] = glCreateProgram();
    glAttachShader(shader_program[POSTPROCESSING_RECT], load_shader(GL_VERTEX_SHADER, "rect_vs.glsl"));
    glAttachShader(shader_program[POSTPROCESSING_RECT], load_shader(GL_FRAGMENT_SHADER, "rect_fs.glsl"));
    link_and_validate_program(shader_program[POSTPROCESSING_RECT]);

    return 1;
}


int init_buffers() {
    glGenVertexArrays(VERTEX_ARRAY_COUNT, vertex_array);

    glBindVertexArray(vertex_array[PLANE]);

    glGenBuffers(1, &vertex_buffer[POSITION]);
    glBindBuffer(GL_ARRAY_BUFFER, vertex_buffer[POSITION]);
    glBufferData(GL_ARRAY_BUFFER, sizeof(fighter_plane), fighter_plane, GL_STATIC_DRAW);
    glVertexAttribPointer(POSITION, 3, GL_FLOAT, GL_FALSE, 0, NULL);

    glGenBuffers(1, &vertex_buffer[TEXTURE_COORD]);
    glBindBuffer(GL_ARRAY_BUFFER, vertex_buffer[TEXTURE_COORD]);
    glBufferData(GL_ARRAY_BUFFER, sizeof(tex_coord_fighter), tex_coord_fighter, GL_STATIC_DRAW);
    glVertexAttribPointer(TEXTURE_COORD, 2, GL_FLOAT, GL_FALSE, 0, NULL);

    glEnableVertexAttribArray(POSITION);
    glEnableVertexAttribArray(TEXTURE_COORD);

    glGenBuffers(1, &vertex_indices_buffer);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vertex_indices_buffer);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(indices), indices, GL_STATIC_DRAW);

    glBindVertexArray(0);



    glBindVertexArray(vertex_array[RECTANGLE]);

    glGenBuffers(1, &vertex_buffer[POSITION_RECT]);
    glBindBuffer(GL_ARRAY_BUFFER, vertex_buffer[POSITION_RECT]);
    glBufferData(GL_ARRAY_BUFFER, sizeof(position_rect), position_rect, GL_STATIC_DRAW);
    glVertexAttribPointer(POSITION, 2, GL_FLOAT, GL_FALSE, 0, NULL);

    glGenBuffers(1, &vertex_buffer[TEXCOORD_RECT]);
    glBindBuffer(GL_ARRAY_BUFFER, vertex_buffer[TEXCOORD_RECT]);
    glBufferData(GL_ARRAY_BUFFER, sizeof(tex_coord_rect), tex_coord_rect, GL_STATIC_DRAW);
    glVertexAttribPointer(TEX_COORD, 2, GL_FLOAT, GL_FALSE, 0, NULL);

    glEnableVertexAttribArray(POSITION);
    glEnableVertexAttribArray(TEX_COORD);

    glBindVertexArray(0);

    return 1;
}

void update() {return;
    unsigned int time_now = SDL_GetTicks();
    if (last_time != 0) {
        unsigned long elapsed = time_now - last_time;
    }
    last_time = time_now;
}

void render() {
    char buf[256];

    glBindFramebuffer(GL_DRAW_FRAMEBUFFER, frameBuffer);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // przepisanie danych w tablicy macierzy modelu-widoku
    for (int layer = FRAME_LAYERS - 1; layer > 0 ; layer--) {
        memcpy(modelview_matrix[layer], modelview_matrix[layer-1], sizeof(float) * 16);
    }


    get_identity_matrix(modelview_matrix[0]);
    //dump_matrix4x4(modelview_matrix[0]);
    get_translation_matrix4x4(aux_matrix1, 0.0f, 0.0f, -(near+far)/2.0f);
    mult_matrices(aux_matrix2, modelview_matrix[0], aux_matrix1, 4);

    // przesunięcie obiektu
    get_translation_matrix4x4(aux_matrix1, translate_x, translate_y, 0.0f);
    mult_matrices(modelview_matrix[0], aux_matrix2, aux_matrix1, 4);

    // włączenie obiektu tablic wierzchołków
    glBindVertexArray(vertex_array[PLANE]);
    glBindTexture(GL_TEXTURE_2D, texture);

    // włączenie programu
    glUseProgram(shader_program[MOTION_BLUR_SHADING]);

    glUniform1i(glGetUniformLocation(shader_program[MOTION_BLUR_SHADING], "tex"), 0);
    // załadowanie zmiennej jednorodnej - iloczynów macierzy modelu-widoku i rzutowania
    for (int layer = 0; layer < FRAME_LAYERS; layer++) {
        sprintf(buf, "modelViewProjectionMatrix[%d]", layer);
        mult_matrices(aux_matrix1, projection_matrix, modelview_matrix[layer], 4);
        transpose_matrix(mvp_matrix, aux_matrix1, 4);
        glUniformMatrix4fv(glGetUniformLocation(shader_program[MOTION_BLUR_SHADING], buf), 1, GL_FALSE, mvp_matrix);
    }

    // narysowanie danych zawartych w tablicach wierzchołków
    glDrawElements(GL_TRIANGLES, 2 * 3, GL_UNSIGNED_INT, NULL);

    // wyłączenie programu
    glUseProgram(0);

    glBindTexture(GL_TEXTURE_2D, 0);

    // wyłączenie obiektu tablic wierzchołków
    glBindVertexArray(0);



    // bufor ramki do zapisu = domyślny bufor ramki
    glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);

    // czyszczenie bufora koloru
    glClear(GL_COLOR_BUFFER_BIT);

    // wyłączenie testu bufora głębokości
    glDisable(GL_DEPTH_TEST);

    // wybór tablicy tekstur 2D
    glBindTexture(GL_TEXTURE_2D_ARRAY, renderTexture[COLOR0]);

    // włączenie obiektu tablic wierzchołków
    glBindVertexArray(vertex_array[RECTANGLE]);

    // włączenie programu
    glUseProgram(shader_program[POSTPROCESSING_RECT]);

    // załadowanie zmiennej jednorodnej - uchwyt tekstury
    glUniform1i(glGetUniformLocation(shader_program[POSTPROCESSING_RECT], "tex"), 0);

    // narysowanie danych zawartych w tablicach wierzchołków
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

    // wyłączenie programu
    glUseProgram(0);

    // wyłączenie obiektu tablic wierzchołków
    glBindVertexArray(0);

    // wyłączenie tablicy tekstur 2D
    glBindTexture(GL_TEXTURE_2D_ARRAY, 0);

    glDisable(GL_DEPTH_TEST);

    draw_text(6.0f, 10.0f, "Naciskaj klawisze strzalek, zeby sterowac mysliwcem", 14.0f, 0.0f, 0.0f, 0.0f, 1.0f);

    glEnable(GL_DEPTH_TEST);
}

void cleanup() {
    cleanup_draw_text();

    if (texture) {
        glDeleteBuffers(1, &texture);
    }
    if (vertex_indices_buffer) {
        glDeleteBuffers(1, &vertex_indices_buffer);
    }
    glDeleteBuffers(VERTEX_BUFFER_COUNT, vertex_buffer);
    glDeleteVertexArrays(VERTEX_ARRAY_COUNT, vertex_array);

    glDeleteProgram(shader_program[MOTION_BLUR_SHADING]);
    glDeleteProgram(shader_program[POSTPROCESSING_RECT]);

    glDeleteTextures(RENDER_TEXTURE_COUNT, renderTexture);

    glDeleteFramebuffers(1, &frameBuffer);

    if (gContext) {
        SDL_GL_DeleteContext(gContext);
        gContext = NULL;
    }

    if (gWindow) {
        SDL_DestroyWindow(gWindow);
        gWindow = NULL;
    }

    IMG_Quit();
    SDL_Quit();
}



int main(int argc, char* args[]) {
    int quit;
    SDL_Event e;

    if (!init()) {
        printf("Błąd inicjalizacji!\n");
        cleanup();
        return 1;
    }

    quit = 0;
    while (!quit) {
        while (SDL_PollEvent(&e) != 0) {
            if (e.type == SDL_QUIT) {
                quit = 1;
                break;
            } else if (e.type == SDL_KEYDOWN) {
                switch (e.key.keysym.sym) {
                    case SDLK_LEFT:
                        translate_x -= 0.03f;
                        break;
                    case SDLK_RIGHT:
                        translate_x += 0.03f;
                        break;
                    case SDLK_UP:
                        translate_y += 0.03f;
                        break;
                    case SDLK_DOWN:
                        translate_y -= 0.03f;
                        break;
                }
            }
        }

        update();
        render();

        SDL_GL_SwapWindow(gWindow);
    }

    cleanup();

    return 0;
}



char *text_file_read(char *fname) {
    FILE *fp;
    char *content = NULL;
    int count = 0;

    if (fname != NULL) {
        fp = fopen(fname, "rt");

        if (fp != NULL) {
            fseek(fp, 0, SEEK_END);
            count = ftell(fp);
            rewind(fp);

            if (count > 0) {
                content = (char*)malloc(sizeof(char) * (count+1));
                count = fread(content,sizeof(char),count,fp);
                content[count] = '\0';
            }
            fclose(fp);
        }
    }
    return content;
}


int check_gl_error() {
    GLenum error;

    error = glGetError();
    switch (error) {
        case GL_CONTEXT_LOST:
            printf("GL_CONTEXT_LOST [%d]: %s\n", error, gluErrorString(error));
            return GL_CONTEXT_LOST;
        case GL_INVALID_ENUM:
            printf("GL_INVALID_ENUM [%d]: %s\n", error, gluErrorString(error));
            return GL_INVALID_ENUM;
        case GL_INVALID_VALUE:
            printf("GL_INVALID_VALUE [%d]: %s\n", error, gluErrorString(error));
            return GL_INVALID_VALUE;
        case GL_INVALID_OPERATION:
            printf("GL_INVALID_OPERATION [%d]: %s\n", error, gluErrorString(error));
            return 1;
        case GL_INVALID_FRAMEBUFFER_OPERATION:
            printf("GL_INVALID_FRAMEBUFFER_OPERATION [%d]: %s\n", error, gluErrorString(error));
            return GL_INVALID_FRAMEBUFFER_OPERATION;
        case GL_OUT_OF_MEMORY:
            printf("GL_OUT_OF_MEMORY [%d]: %s\n", error, gluErrorString(error));
            return 1;
        case GL_STACK_OVERFLOW:
            printf("GL_STACK_OVERFLOW [%d]: %s\n", error, gluErrorString(error));
            return GL_STACK_OVERFLOW;
        case GL_STACK_UNDERFLOW:
            printf("GL_STACK_UNDERFLOW [%d]: %s\n", error, gluErrorString(error));
            return GL_STACK_UNDERFLOW;
        case GL_NO_ERROR:
            printf("GL_NO_ERROR [%d]\n", error);
            break;
    }

    return GL_NO_ERROR;
}
